__title__ = "dbus_fast"
__description__ = (
    "A performant zero-dependency DBus library for Python with asyncio support"
)
__url__ = "https://github.com/bluetooth-devices/dbus-fast"
__version__ = "4.3.0"
__author__ = "Bluetooth Devices authors, Tony Crisci"
__author_email__ = "bluetooth@koston.org"
__license__ = "MIT"
__copyright__ = "Copyright 2022 Bluetooth Devices authors, 2019 Tony Crisci"
